@@ -1,4 +1,5 @@
-from functools import wraps
+import math
+from functools import wraps, partial
 
 import torch
 import torch.nn.functional as F
@@ -23,6 +24,9 @@ from beartype import beartype
 def exists(val):
     return val is not None
 
+def first(it):
+    return it[0]
+
 def default(val, d):
     return val if exists(val) else d
 
@@ -55,6 +59,16 @@ def log(t, eps = 1e-20):
 def l2norm(t):
     return F.normalize(t, p = 2, dim = -1)
 
+def matrix_diag(t):
+    device = t.device
+    i, j = t.shape[-2:]
+    num_diag_el = min(i, j)
+    i_range = torch.arange(i, device = device)
+    j_range = torch.arange(j, device = device)
+    diag_mask = rearrange(i_range, 'i -> i 1') == rearrange(j_range, 'j -> 1 j')
+    diag_el = t.masked_select(diag_mask)
+    return rearrange(diag_el, '(b d) -> b d', d = num_diag_el)
+
 # 2d sinusoidal positional embedding
 # simple vit paper shows it is good enough compared to learned
 
@@ -78,13 +92,15 @@ def posemb_sincos_2d(patches, temperature = 10000, dtype = torch.float32):
 # biasless layernorm
 
 class LayerNorm(nn.Module):
-    def __init__(self, dim):
+    def __init__(self, dim, scale = True):
         super().__init__()
-        self.gamma = nn.Parameter(torch.ones(dim))
-        self.register_buffer('beta', torch.zeros(dim))
+        self.learned_gamma = nn.Parameter(torch.ones(dim)) if scale else None
+
+        self.register_buffer('gamma', torch.ones(dim), persistent = False)
+        self.register_buffer('beta', torch.zeros(dim), persistent = False)
 
     def forward(self, x):
-        return F.layer_norm(x, x.shape[-1:], self.gamma, self.beta)
+        return F.layer_norm(x, x.shape[-1:], default(self.learned_gamma, self.gamma), self.beta)
 
 # feedforward
 
@@ -218,14 +234,90 @@ class Transformer(nn.Module):
         self,
         x,
         rel_pos_bias = None,
-        mask = None
+        mask = None,
+        return_all_layers = False
     ):
+        layers = []
 
         for attn, ff in self.layers:
             x = attn(x, rel_pos_bias = rel_pos_bias, mask = mask) + x
             x = ff(x) + x
+            layers.append(x)
 
-        return x
+        if not return_all_layers:
+            return x
+
+        return x, torch.stack(layers[:-1])
+
+# contrastive losses
+
+class SoftmaxContrastiveLearning(nn.Module):
+    def __init__(
+        self,
+        *,
+        layers = 1,
+        decoupled_contrastive_learning = False,
+        init_temp = 10
+    ):
+        super().__init__()
+        self.temperatures = nn.Parameter(torch.ones(layers, 1, 1) * math.log(init_temp))
+        self.decoupled_contrastive_learning = decoupled_contrastive_learning
+
+    @property
+    def device(self):
+        return next(self.parameters()).device
+
+    def forward(self, sims):
+        batch = sims.shape[-1]
+
+        if sims.ndim == 2:
+            sims = rearrange(sims, 'i j -> 1 i j')
+
+        sims = sims * self.temperatures.exp()
+
+        cosine_sims_exp = sims.exp()
+
+        numerator = matrix_diag(cosine_sims_exp)
+
+        if self.decoupled_contrastive_learning:
+            eye = torch.eye(batch, device = self.device, dtype = torch.bool)
+            cosine_sims_exp = cosine_sims_exp.masked_fill(eye, 0.)
+
+        denominator_i = reduce(cosine_sims_exp, 'l i j -> l i', 'sum')
+        denominator_j = reduce(cosine_sims_exp, 'l i j -> l j', 'sum')
+
+        contrastive_loss = -log(numerator) + 0.5 * (log(denominator_i) + log(denominator_j))
+
+        contrastive_loss = reduce(contrastive_loss, 'l n -> l', 'mean')
+        return contrastive_loss.sum()
+
+class SigmoidContrastiveLearning(nn.Module):
+    """ https://arxiv.org/abs/2303.15343 """
+
+    def __init__(
+        self,
+        *,
+        layers = 1,
+        init_temp = 10,
+        init_bias = -10
+    ):
+        super().__init__()
+        self.temperatures = nn.Parameter(torch.ones(layers, 1, 1) * math.log(init_temp))
+        self.bias = nn.Parameter(torch.ones(layers, 1, 1) * init_bias)
+
+    @property
+    def device(self):
+        return next(self.parameters()).device
+
+    def forward(self, sims):
+        if sims.ndim == 2:
+            sims = rearrange(sims, 'i j -> 1 i j')
+
+        n = sims.shape[-1]
+        sims = sims * self.temperatures.exp() + self.bias
+        labels = 2 * rearrange(torch.eye(n), 'i j -> 1 i j') - torch.ones_like(sims)
+
+        return -F.logsigmoid(labels * sims).sum() / n
 
 # Audio Spectrogram Transformer - https://arxiv.org/abs/2104.01778
 
@@ -243,6 +335,8 @@ class AudioSpectrogramTransformer(nn.Module):
         attn_dropout = 0.,
         ff_mult = 4,
         ff_dropout = 0.,
+        accept_spec = False,
+        accept_spec_time_first = True,
         spec_n_fft = 128,
         spec_power = 2,
         spec_win_length = 24,
@@ -257,6 +351,7 @@ class AudioSpectrogramTransformer(nn.Module):
     ):
         super().__init__()
         self.dim = dim
+        self.depth = depth
 
         self.patch_size = pair(patch_size)
         patch_input_dim = self.patch_size[0] * self.patch_size[1]
@@ -268,6 +363,9 @@ class AudioSpectrogramTransformer(nn.Module):
             nn.LayerNorm(dim)
         )
 
+        self.accept_spec = accept_spec
+        self.accept_spec_time_first = accept_spec_time_first
+
         self.spec = Spectrogram(
             n_fft = spec_n_fft,
             power = spec_power,
@@ -318,11 +416,17 @@ class AudioSpectrogramTransformer(nn.Module):
     def forward(
         self,
         x,
-        force_no_patch_dropout = False
+        force_no_patch_dropout = False,
+        return_all_layers = False
     ):
         batch, device = x.shape[0], x.device
+        assert (self.accept_spec and x.ndim == 3) or (not self.accept_spec and x.ndim == 2)
 
-        x = self.spec(x)
+        if self.accept_spec and self.accept_spec_time_first:
+            x = rearrange(x, 'b t f -> b f t')
+
+        if not self.accept_spec:
+            x = self.spec(x)
 
         if self.training:
             x = self.aug(x)
@@ -384,13 +488,18 @@ class AudioSpectrogramTransformer(nn.Module):
 
         # attention, what else
 
-        x = self.transformer(x, rel_pos_bias = rel_pos_bias)
+        x, all_layers = self.transformer(x, rel_pos_bias = rel_pos_bias, return_all_layers = True)
 
         # final global average and norm (most recent papers show this is superior to CLS token)
 
         x = reduce(x, 'b n d -> b d', 'mean')
 
-        return self.norm(x)
+        out = self.norm(x)
+
+        if not return_all_layers:
+            return out
+
+        return out, all_layers
 
 # text transformer
 
@@ -415,6 +524,7 @@ class TextTransformer(nn.Module):
         self.token_emb = nn.Embedding(num_tokens, dim)
         self.pos_emb = nn.Embedding(max_seq_len, dim)
 
+        self.depth = depth
         self.max_seq_len = max_seq_len
 
         self.cls_token = nn.Parameter(torch.randn(dim))
@@ -440,7 +550,8 @@ class TextTransformer(nn.Module):
         self,
         x = None,
         raw_texts: Optional[List[str]] = None,
-        mask = None
+        mask = None,
+        return_all_layers = False
     ):
         assert exists(x) ^ exists(raw_texts)
 
@@ -471,13 +582,69 @@ class TextTransformer(nn.Module):
 
         # attention
 
-        x = self.transformer(x, mask = mask)
+        x, all_layers = self.transformer(x, mask = mask, return_all_layers = True)
 
         # unpack the cls tokens
 
         cls_tokens, _ = unpack(x, ps, 'b * d')
 
-        return self.norm(cls_tokens)
+        out = self.norm(cls_tokens)
+
+        if not return_all_layers:
+            return out
+
+        return out, all_layers
+
+# hierarchical cl loss
+
+def interspersed_indices(layers, total_layers):
+    assert total_layers >= layers
+    step = total_layers / layers
+    return (torch.arange(0, layers) * step).floor().long()
+
+class MultiLayerContrastiveLoss(nn.Module):
+    def __init__(
+        self,
+        *,
+        audio_dim,
+        text_dim,
+        dim_latent,
+        layers,
+        decoupled_contrastive_learning = False,
+        sigmoid_contrastive_loss = False
+    ):
+        super().__init__()
+        self.layers = layers
+
+        self.audio_norm = LayerNorm(audio_dim, scale = False)
+        self.audio_gamma = nn.Parameter(torch.ones(layers, 1, audio_dim))
+        self.audio_latent_weight = nn.Parameter(torch.randn(layers, audio_dim, dim_latent))
+        self.audio_latent_bias = nn.Parameter(torch.randn(layers, 1, dim_latent))
+
+        self.text_norm = LayerNorm(text_dim, scale = False)
+        self.text_gamma = nn.Parameter(torch.ones(layers, 1, text_dim))
+        self.text_latent_weight = nn.Parameter(torch.randn(layers, text_dim, dim_latent))
+        self.text_latent_bias = nn.Parameter(torch.randn(layers, 1, dim_latent))
+
+        klass = SigmoidContrastiveLearning if sigmoid_contrastive_loss else partial(SoftmaxContrastiveLearning, decoupled_contrastive_learning = decoupled_contrastive_learning)
+        self.contrast = klass(layers = layers)
+
+    def forward(self, *, audio_layers, text_layers):
+        device, batch = audio_layers.device, audio_layers.shape[1]
+
+        audio_gap = reduce(audio_layers, 'l b n d -> l b d', 'mean')
+        audio_embeds = self.audio_norm(audio_gap) * self.audio_gamma
+        audio_latents = einsum('l b d, l d e -> l b e', audio_embeds, self.audio_latent_weight) + self.audio_latent_bias
+        audio_latents = l2norm(audio_latents)
+
+        text_cls_tokens = text_layers[:, :, 0]
+        text_embeds = self.text_norm(text_cls_tokens) * self.text_gamma
+        text_latents = einsum('l b d, l d e -> l b e', text_embeds, self.text_latent_weight) + self.text_latent_bias
+        text_latents = l2norm(text_latents)
+
+        cosine_sims = einsum('l i d, l j d -> l i j', audio_latents, text_latents)
+
+        return self.contrast(cosine_sims)
 
 # main classes
 
@@ -489,6 +656,9 @@ class MuLaN(nn.Module):
         text_transformer: TextTransformer,
         dim_latent = 128,                       # they use 128
         decoupled_contrastive_learning = True,  # think this was used, make it optional
+        hierarchical_contrastive_loss = False,
+        hierarchical_contrastive_loss_layers = None,
+        sigmoid_contrastive_loss = False
     ):
         super().__init__()
         self.dim_latent = dim_latent
@@ -496,63 +666,103 @@ class MuLaN(nn.Module):
         self.audio = audio_transformer
         self.text = text_transformer
 
-        self.temperature = nn.Parameter(torch.tensor(1.))
 
         self.text_to_latents = nn.Linear(self.text.dim, dim_latent)
         self.audio_to_latents = nn.Linear(self.audio.dim, dim_latent)
 
-        self.decoupled_contrastive_learning = decoupled_contrastive_learning
+        klass = SigmoidContrastiveLearning if sigmoid_contrastive_loss else partial(SoftmaxContrastiveLearning, decoupled_contrastive_learning = decoupled_contrastive_learning)
+        self.contrast = klass()
+
+        self.multi_layer_contrastive_learning = None
+
+        if hierarchical_contrastive_loss:
+            num_layers = default(hierarchical_contrastive_loss_layers, min(audio_transformer.depth, text_transformer.depth) - 1)
+            assert num_layers > 0
+
+            self.register_buffer('text_layers_indices', interspersed_indices(num_layers, text_transformer.depth))
+            self.register_buffer('audio_layers_indices', interspersed_indices(num_layers, audio_transformer.depth))
+
+            self.multi_layer_contrastive_learning = MultiLayerContrastiveLoss(
+                audio_dim = self.audio.dim,
+                text_dim = self.text.dim,
+                dim_latent = dim_latent,
+                layers = num_layers,
+                decoupled_contrastive_learning = decoupled_contrastive_learning,
+                sigmoid_contrastive_loss = sigmoid_contrastive_loss
+            )
 
     def get_audio_latents(
         self,
-        wavs
+        wavs,
+        return_all_layers = False
     ):
-        audio_embeds = self.audio(wavs)
+        audio_embeds, audio_layers = self.audio(wavs, return_all_layers = True)
         audio_latents = self.audio_to_latents(audio_embeds)
-        return l2norm(audio_latents)
+        out = l2norm(audio_latents)
+
+        if not return_all_layers:
+            return out
+
+        return out, audio_layers
 
     def get_text_latents(
         self,
         texts = None,
-        raw_texts: Optional[List[str]] = None
+        raw_texts: Optional[List[str]] = None,
+        return_all_layers = False
     ):
-        text_embeds = self.text(texts, raw_texts = raw_texts)
+        text_embeds, text_layers = self.text(texts, raw_texts = raw_texts, return_all_layers = True)
         text_latents = self.text_to_latents(text_embeds)
-        return l2norm(text_latents)
+        out = l2norm(text_latents)
+
+        if not return_all_layers:
+            return out
+
+        return out, text_layers
 
     def forward(
         self,
         wavs,
         texts = None,
         raw_texts: Optional[List[str]] = None,
-        return_similarities = False
+        return_latents = False,
+        return_similarities = False,
+        return_pairwise_similarities = False
     ):
         batch, device = wavs.shape[0], wavs.device
 
-        audio_latents = self.get_audio_latents(wavs)
-        text_latents = self.get_text_latents(texts, raw_texts = raw_texts)
+        audio_latents, audio_layers = self.get_audio_latents(wavs, return_all_layers = True)
+        text_latents, text_layers = self.get_text_latents(texts, raw_texts = raw_texts, return_all_layers = True)
+
+        if return_latents:
+            return audio_latents, text_latents
+
+        if return_similarities:
+            return einsum('i d, i d -> i', audio_latents, text_latents)
 
         cosine_sim = einsum('i d, j d -> i j', audio_latents, text_latents)
 
         assert cosine_sim.shape[0] == cosine_sim.shape[1], 'batch sizes for audio and text are not equal'
 
-        if return_similarities:
+        if return_pairwise_similarities:
             return cosine_sim
 
-        cosine_sim = cosine_sim * self.temperature.exp()
+        cl_loss = self.contrast(cosine_sim)
 
-        cosine_sim_exp = cosine_sim.exp()
+        if not exists(self.multi_layer_contrastive_learning):
+            return cl_loss
 
-        numerator = cosine_sim_exp.diag()
+        audio_layers = audio_layers[self.audio_layers_indices]
+        text_layers = text_layers[self.text_layers_indices]
 
-        if self.decoupled_contrastive_learning:
-            eye = torch.eye(batch, device = device, dtype = torch.bool)
-            cosine_sim_exp = cosine_sim_exp.masked_fill(eye, 0.)
+        # whether to do cl loss across all layers, from ViCHA paper https://arxiv.org/abs/2208.13628
 
-        denominator = reduce(cosine_sim_exp, 'i j -> i', 'sum')
+        hierarchical_cl_loss = self.multi_layer_contrastive_learning(
+            audio_layers = audio_layers,
+            text_layers = text_layers
+        )
 
-        contrastive_loss = -log(numerator) + log(denominator)
-        return contrastive_loss.mean()
+        return cl_loss + hierarchical_cl_loss
 
 # music lm
 
@@ -661,13 +871,34 @@ class MusicLM(nn.Module):
     @torch.no_grad()
     def forward(
         self,
-        raw_texts: List[str],
+        text: str,
+        num_samples = 1,
         **audio_lm_kwargs
     ):
         self.eval()
 
-        texts = tokenizer.tokenize(raw_texts).to(self.device)
+        texts = tokenizer.tokenize([text]).to(self.device)
 
         text_embeds = self.mulan_embed_quantizer(texts = texts)
 
-        return self.audio_lm(text_embeds = text_embeds, **audio_lm_kwargs)
+        # unable to deal with variable lengthed audio for now
+
+        samples = []
+
+        for _ in range(num_samples):
+            music = self.audio_lm(text_embeds = text_embeds, **audio_lm_kwargs)
+            samples.append(music)
+
+        # if one sample, just return it
+
+        if num_samples == 1:
+            return first(samples)
+
+        mulan = self.mulan_embed_quantizer.mulan
+
+        # get the one with the highest similarity score, of all the samples
+
+        sims = torch.cat([mulan(texts = texts, wavs = music, return_similarities = True) for music in samples], dim = 0)
+        top_matching_index = sims.topk(1, dim = 0).indices.item()
+
+        return samples[top_matching_index]
